#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/poll.h>
#include <sys/mman.h>

#include <netinet/in.h>

#include <linux/if_ether.h>
#include <linux/tcp.h>
#include <linux/ip.h>
#include <linux/logdev.h>

#include "dump_log.h"

#define LOGDEV_FILE "/dev/logdev"
#define LOGDEV_PROC "/proc/logdev"
#define LOGDEV_SWITCH "/proc/logdev/switch"
#define LOGDEV_MINOR "minor"
#define MISC_MAJOR 10

/* todo, make this generic! */
#define PAGE_SHIFT 12
#define PAGE_SIZE (1<<PAGE_SHIFT)
#define PAGE_MASK (~(PAGE_SIZE-1))

static int last_id;

static char *logdev_version = "0.1.0";

char *dump_log_version(void)
{
	return logdev_version;
}

void process_switch(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_switch_struct *rs = &item->u.sw;
	char *pcomm;
	char *ncomm;
	unsigned long long t;
	unsigned long secs;
	unsigned long nsecs;

	t = (unsigned long long)rs->high << 32 | rs->low;
	secs = (unsigned long)(t / 1000000000ULL);
	nsecs = (unsigned long)(t % 1000000000ULL);
	
	pcomm = malloc(rs->prev_len+1);
	if (!pcomm) {
		perror("malloc");
		exit(-1);
	}
	ncomm = malloc(rs->next_len+1);
	if (!ncomm) {
		perror("malloc");
		exit(-1);
	}
	memcpy(pcomm,&rs->prev_comm[0],rs->prev_len);
	memcpy(ncomm,&rs->prev_comm[rs->prev_len],rs->next_len);

	pcomm[rs->prev_len]=0;
	ncomm[rs->next_len]=0;
	printf("[%5lu.%06lu] CPU:%d (%s:%d) -->> (%s:%d)\n",
	       secs,
	       nsecs/1000,
	       hdr->cpu,
	       pcomm,rs->pid_prev,
	       ncomm,rs->pid_next);
	free(pcomm);
	free(ncomm);
}

#define IPBYTE(ip,b) (unsigned char)(((ip)>>((b)*8))&0xff)

void process_pkt(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_pkt *rs = &item->u.pkt;
	unsigned long end;
	struct tcphdr *th;
	struct iphdr *iph;
	unsigned long oaddr;
	unsigned long iaddr;
	unsigned short oport;
	unsigned short iport;
	int iplen;
	int tcplen;
	int len;
	int dir;
	unsigned long long t;
	unsigned long secs;
	unsigned long nsecs;

	t = (unsigned long long)rs->high << 32 | rs->low;
	secs = (unsigned long)(t / 1000000000ULL);
	nsecs = (unsigned long)(t % 1000000000ULL);
	

	printf("[%5lu.%06lu] CPU:%d ",
	       secs,
	       nsecs/1000,
	       hdr->cpu);

	if (rs->protocol != htons(ETH_P_IP)) {
		printf("Not IP packet\n");
		return;
	}
	
	dir = rs->dir;

	iph = (struct iphdr*)rs->packet;
	iph->tot_len = ntohs(iph->tot_len);
	iph->id = ntohs(iph->id);
	iph->frag_off = ntohs(iph->frag_off);
	iph->check = ntohs(iph->check);
	iph->saddr = ntohl(iph->saddr);
	iph->daddr = ntohl(iph->daddr);
	
	if (dir) {
		/* outgoing */
		oaddr = iph->daddr;
		iaddr = iph->saddr;
	} else {
		/* incomming */
		oaddr = iph->saddr;
		iaddr = iph->daddr;
	}
	
	if (iph->protocol != 6) {
		printf("%d.%d.%d.%d %s %d.%d.%d.%d (Not TCP packet)\n",
		       IPBYTE(iaddr,3),
		       IPBYTE(iaddr,2),
		       IPBYTE(iaddr,1),
		       IPBYTE(iaddr,0),
		       dir ? "==>" : "<==",
		       IPBYTE(oaddr,3),
		       IPBYTE(oaddr,2),
		       IPBYTE(oaddr,1),
		       IPBYTE(oaddr,0));
		return;
	}

	iplen = iph->ihl<<2;

	th = (struct tcphdr*)((void*)(iph)+(iplen));
	tcplen = th->doff<<2;

	th->source = ntohs(th->source);
	th->dest = ntohs(th->dest);
	th->seq = ntohl(th->seq);
	th->ack_seq = ntohl(th->ack_seq);
	th->window = ntohs(th->window);
	th->check = ntohs(th->check);
	th->urg_ptr = ntohs(th->urg_ptr);

	len = iph->tot_len - (iplen + tcplen);
	end = th->seq;
	if (th->fin || th->syn) {
		end++;
	}
	end += len;

	if (dir == 3)
		printf("dequeue: ");
	
	if (dir) {
		/* outgoing */
		oport = th->dest;
		iport = th->source;
	} else {
		/* incomming */
		oport = th->source;
		iport = th->dest;
	}


	printf("%d.%d.%d.%d:%d %s %d.%d.%d.%d:%d seq:%u ack:%u "
	       "(%c%c%c%c%c%c) len:%d win:%d end_seq:%lu\n",
	       IPBYTE(iaddr,3),
	       IPBYTE(iaddr,2),
	       IPBYTE(iaddr,1),
	       IPBYTE(iaddr,0),
	       iport,
	       dir ? "==>" : "<==",
	       IPBYTE(oaddr,3),
	       IPBYTE(oaddr,2),
	       IPBYTE(oaddr,1),
	       IPBYTE(oaddr,0),
	       oport,
	       th->seq,
	       th->ack_seq,
	       th->fin ? 'F' : '-',
	       th->syn ? 'S' : '-',
	       th->rst ? 'R' : '-',
	       th->psh ? 'P' : '-',
	       th->ack ? 'A' : '-',
	       th->urg ? 'U' : '-',
	       len,
	       th->window,
	       end
	       );
}

void process_print(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_print *rs = &item->u.print;
	char *buf;
	int len;

	len = hdr->size - (sizeof(*rs) + sizeof(*hdr));
	buf = malloc(len+1);
	if (!buf) {
		perror("malloc");
		exit(-1);
	}
	strncpy(buf,rs->str,len);
	buf[len] = 0;

	if (last_id != LOGDEV_PRINT)
		printf("print:  ");
	printf("  %s", buf);
	free (buf);
}

void process_print_time(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_print_time *rs = &item->u.print_time;
	char *buf;
	int len;
	unsigned long long t;
	unsigned long secs;
	unsigned long nsecs;

	t = (unsigned long long)rs->high << 32 | rs->low;
	secs = (unsigned long)(t / 1000000000ULL);
	nsecs = (unsigned long)(t % 1000000000ULL);
	

	len = hdr->size - (sizeof(*rs) + sizeof(*hdr));
	buf = malloc(len+1);
	if (!buf) {
		perror("malloc");
		exit(-1);
	}
	strncpy(buf,rs->str,len);
	buf[len] = 0;

	if (last_id != LOGDEV_PRINT)
		printf("print: [%5lu.%06lu]",
				secs,
				nsecs);
	printf("  %s", buf);
	free (buf);
}

void process_print_time_func(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_print_time_func *rs = &item->u.print_time_func;
	char *buf;
	int len;
	unsigned long long t;
	unsigned long secs;
	unsigned long nsecs;
	static int kfd = -1;

	if (kfd == -1)
		kfd = open("/dev/kmem",O_RDONLY);
		
	t = (unsigned long long)rs->high << 32 | rs->low;
	secs = (unsigned long)(t / 1000000000ULL);
	nsecs = (unsigned long)(t % 1000000000ULL);
	

	len = hdr->size - (sizeof(*rs) + sizeof(*hdr));
	buf = malloc(len+1);
	if (!buf) {
		perror("malloc");
		exit(-1);
	}
	strncpy(buf,rs->str,len);
	buf[len] = 0;

	if (last_id != LOGDEV_PRINT) {
		printf("print: [%5lu.%06lu] ",
				secs,
				nsecs);
		if (kfd >= 0) {
			off_t ptr = (off_t)rs->file & ~(PAGE_MASK);
			off_t offset = (off_t)rs->file & PAGE_MASK;
			int i = 0;
			char *map;

again:
			map = mmap(NULL,PAGE_SIZE,PROT_READ,MAP_SHARED,kfd,offset);
			if (map == MAP_FAILED) {
				printf("(map_failed:%p):%d ",
						rs->file,
						rs->line);
			} else {
				char str[PAGE_SIZE];
				while (ptr < PAGE_SIZE && i < PAGE_SIZE) {
					if (!map[ptr])
						break;
					str[i++] = map[ptr++];
				}
				if (ptr == PAGE_SIZE && i < PAGE_SIZE) {
					munmap(map,PAGE_SIZE);
					offset += PAGE_SIZE;
					ptr = 0;
					goto again;
				}
				if (i >= PAGE_SIZE)
					str[PAGE_SIZE-1] = 0;
				else
					str[i] = 0;
				printf("%s:%d ",
						str,
						rs->line);
				munmap(map,PAGE_SIZE);
			}
		}
	}
			
	printf("  %s", buf);
	free (buf);
}

#ifdef HAVE_RESOURCE_SETS
typedef int tracer_t;
#include <../../resource_sets/dump_trace.h>
#endif
void process_custom(struct logdev_item *item)
{
	struct logdev_header *hdr = &item->hdr;
	struct logdev_custom *rs = &item->u.custom;
	char *buf = rs->data;
	int size;
	int i;
	int x = 0;

	size = hdr->size - ((sizeof(*hdr)) + sizeof(*rs));

	switch (rs->id) {
#ifdef HAVE_RESOURCE_SETS
	case 0xabb:
	{
		struct logdev_resource_set_switch *rs = dat;
		
		printf("custom ...\n");
		printf("  RS_SWITCH:\n");
		if (rs->next_crrs) {
			printf("    time: %d:%06d\n"
			       "    cpu: %d\n"
			       "    prevrs: %d\n"
			       "    nextrs: %d\n"
			       "    deadline: %d:%06d\n"
			       "    period: %d:%06d\n"
			       "    future = %d\n",
			       (unsigned int)rs->tv.tv_sec,(int)rs->tv.tv_usec,
			       rs->cpu,
			       rs->prev_crrs,
			       rs->next_crrs,
			       (unsigned int)rs->next_deadline.tv_sec,(int)rs->next_deadline.tv_usec,
			       (unsigned int)rs->next_period.tv_sec,(int)rs->next_period.tv_usec,
			       rs->future);
		} else {
			printf("    time: %d:%06d\n"
			       "    cpu: %d\n"
			       "    prevrs: %d\n"
			       "    nextrs: (IDLE)\n",
			       (unsigned int)rs->tv.tv_sec,(int)rs->tv.tv_usec,
			       rs->cpu,
			       rs->prev_crrs);
		}
		break;
	}
	case 0xabc:
	case 0xabd:
#endif
	default:

		printf("custom size=%d  custom_id=%x",hdr->size,rs->id);
		for (i=0; i < size; i++) {
			if (!(x & 15)) {
				printf("\n   ");
			} else if (!(x & 7)) {
				printf(" ");
			}
			x++;
			printf("%02x ",buf[i]);
		}
		printf("\n");
	}


}

struct sys_rec {
	struct sys_rec *next;
	unsigned long pc;
	char *name;
};

static struct sys_rec *sysnames = NULL;
static struct sys_rec **sysnamesarr;
static int sys_cnt = 0;

static int sys_rec_compare(const void *A, const void *B)
{
	struct sys_rec **recA = (struct sys_rec**)A;
	struct sys_rec **recB = (struct sys_rec**)B;
	struct sys_rec *ra = *recA;
	struct sys_rec *rb = *recB;


	if (ra->pc < rb->pc)
		return -1;
	if ((ra->pc > rb->pc) && rb->next && (ra->pc > rb->next->pc))
		return 1;
	return 0;
}

void load_sys(FILE *fp)
{
	unsigned long pc;
	char buf[100];
	struct sys_rec *ptr;
	struct sys_rec **last = &sysnames;
	int i;

	for (;;) {
		if (fscanf(fp,"%lx %*c %99s\n",&pc,buf)<=0)
			break;
		ptr = malloc(sizeof(*ptr));
		ptr->pc = pc;
		ptr->name = malloc(strlen(buf)+1);
		strcpy(ptr->name,buf);
		sys_cnt++;
		*last = ptr;
		last = &ptr->next;
	}
	ptr->next = NULL;

	sysnamesarr = malloc(sizeof(*sysnamesarr)*sys_cnt);
	for (i=0,ptr=sysnames; ptr ; i++, ptr=ptr->next) {
		sysnamesarr[i] = ptr;
	}
	
	
}
	
char *find_sys(unsigned long key)
{
	struct sys_rec rec;
	struct sys_rec *recP = &rec;
	struct sys_rec **recPP;

	recP->pc = key;
	recPP = bsearch(&recP,sysnamesarr,sys_cnt,sizeof(*sysnamesarr),
			sys_rec_compare);
	if (!recPP)
		return NULL;
	recP = *recPP;
	return recP->name;
}

struct logdev_item *read_dump(int fd, int *id)
{
	struct logdev_header hdr;
	struct logdev_item *buf = NULL;
	int rc;

	do {
#if 0
		struct pollfd logdev_pollfd;
		logdev_pollfd.fd = fd;
		logdev_pollfd.events = POLLIN;
		rc = poll(&logdev_pollfd, 1, -1);
		if (rc < 0) {
			if (errno != EINTR) {
				printf("read_dump: poll error %s\n",strerror(errno));
				exit(888);
			}
			printf("read_dump: poll warning %s\n",strerror(errno));
			return NULL;
		}
#endif

		if ((rc = read(fd,&hdr,sizeof(hdr)))<0) {
			if (errno == EAGAIN) {
				printf("got EAGAIN\n");
				continue;
			}
			perror("read");
			printf("error reading hdr, returning NULL\n");
			return NULL;
		}
		if (!rc)
			return NULL;

		if (LOGDEV_VALID_ID(hdr.id))
			break;
	} while (1);

	buf = malloc(hdr.size);
	if (!buf) {
		printf("ID=%08x size=%d\n",
		       hdr.id,
		       hdr.size);
		perror("OOM");
		exit(-1);
	}

	memcpy(buf,&hdr,sizeof(hdr));
	if (read(fd,buf->u.data,hdr.size - sizeof(hdr))<0) {
		perror("read");
		free(buf);
		return NULL;
	}
	*id = hdr.id;

	return buf;
}

struct logdev_item *read_record(int fd)
{
	struct logdev_header hdr;
	struct logdev_item *rec = NULL;
	int rc;

	if ((rc = read(fd,&hdr,sizeof(hdr))) < sizeof(hdr)) {
		return NULL;
	}

	if (LOGDEV_VALID_ID(hdr.id)) {
		rec = malloc(hdr.size);

		/*
		 * We don't check for error here, since the process
		 * gets what it gets, and it can figure out if it is 
		 * an error or not.
		 */
		memcpy(rec,&hdr,sizeof(hdr));
		read(fd,rec->u.data,hdr.size - sizeof(hdr));
	}

	return rec;
}

void free_record(struct logdev_item *rec)
{
	free(rec);
}

int process_dump(int fd)
{
	int id;
	struct logdev_item *buf;
	
	if ((buf = read_dump(fd,&id)) == 0)
		return 0;

	switch (id) {
	case LOGDEV_SWITCH_ID:
		process_switch(buf);
		break;
	case LOGDEV_PKT_ID:
		process_pkt(buf);
		break;
	case LOGDEV_PRINT:
		process_print(buf);
		break;
	case LOGDEV_PRINT_TIME:
		process_print_time(buf);
		break;
	case LOGDEV_PRINT_TIME_FUNC:
		process_print_time_func(buf);
		break;
	case LOGDEV_CUSTOM:
	{
		process_custom(buf);
		break;
	}
	}
	last_id = id;

	free(buf);

	return 1;
}

int open_logdev_cpu(char *uname, mode_t mode, int cpu)
{
	int fd;
	int r;
	int s;
	int minor;
	char buf[100];
	char *name = LOGDEV_FILE;
	struct stat st;
	dev_t dev;
	mode_t dev_mode;

	if (uname)
		name = uname;

	if ((s = stat(name,&st)) < 0) {
		if (errno != ENOENT)
			return s;
	}

	/*
	 * We could be reading a dump file of the device, so
	 * if this is just a regular file, open it normally.
	 */
	if (!s && S_ISREG(st.st_mode)) {
		return open(uname,mode);
	}

	sprintf(buf,"%s/%d/%s",LOGDEV_PROC,cpu,LOGDEV_MINOR);

	/*
	 * Search the minor numbers for each CPU and see if 
	 * this currently matches, otherwise we just open
	 * the first one.
	 */
	if ((fd = open(buf,O_RDONLY)) < 0)
		return fd;

	/* More than enough! */
	if ((r = read(fd,buf,100)) < 0) 
		return r;

	if (r >= 100) /* ??? */
		r = 99;
	buf[r] = 0;
	minor = atoi(buf);

	if (s || (st.st_dev & 0xff) != minor) {
		if (!s && (r = unlink(name)) < 0)
			return r;
		dev_mode = 0666 | S_IFCHR;
		dev = (MISC_MAJOR << 8)|(minor & 0xff);
		if ((r = mknod(name,dev_mode,dev)) < 0)
			return r;
	}

	return open(name,mode);
}

int open_logdev(char *uname, mode_t mode)
{
	return open_logdev_cpu(uname, mode, 0);
}

int close_logdev(int fd)
{
	return close(fd);
}

int logdev_switch_set(int on)
{
	int fd;
	int r;
	char buf[100];

	if ((fd = open(LOGDEV_SWITCH,O_WRONLY)) < 0)
		return fd;
	
	r = snprintf(buf,99,"%d",on);
	if (r >= 100) /* ??? */
		r = 99;
	buf[r] = 0;

	write(fd,buf,r);

	close(fd);
	
	return 0;
}

int logdev_switch_read(void)
{
	int fd;
	int r;
	char buf[100];

	if ((fd = open(LOGDEV_SWITCH,O_RDONLY)) < 0)
		return fd;
	
	read(fd,buf,100);

	close(fd);

	r = atoi(buf);

	return r;
}

/*
 * This is used for writing to the log device without corrupting it.
 */
int logdev_print(int fd,const char *fmt, ...)
{
	va_list ap;
	char buffer[BUFSIZ];
	int len;
	int free_buf = 0;
	struct logdev_header *hdr;
	struct logdev_print *rs;
	char *buf = buffer;
	
	len = sizeof(*hdr) + sizeof(*rs);

	va_start(ap,fmt);
	len += vsnprintf(buf+len,BUFSIZ-len,fmt,ap);
	va_end(ap);
	if (len >= BUFSIZ) {
		int sz = len+1;
		/* try again! */
		buf = malloc(sz);
		len = sizeof(*hdr) + sizeof(*rs);
		va_start(ap,fmt);
		len += vsnprintf(buf+len,sz,fmt,ap);
		va_end(ap);
		free_buf = 1;
	}
	
	hdr = (struct logdev_header *)buf;
	rs = (struct logdev_print *)(buf + sizeof(*hdr));
	
	hdr->id = LOGDEV_PRINT;
	hdr->size = len;
	hdr->cpu = 0; /* ??? */
	write(fd,buf,len);
	
	if (free_buf)
		free(buf);
	
	return len - (sizeof(*hdr) + sizeof(*rs));
}
	
